#!/usr/bin/env python
# -*- coding: utf-8 -*-
__all__ =["Web","log"]
"""
=======================================================================================
AEWEB - "app engine web" ©2009 manatlan
=======================================================================================
Simple wrapper around webapp/gae
more info on http://code.google.com/appengine/docs/python/tools/webapp/ (redirect() ...)

features :
  - methods return the output body (and can be generators(can use yield to output text))
  - restul operations thru GET/POST/ ... (like webpy)
  - route according http_host (by managing "exception hosts")
  - delegate() method, to dispatch easily -> "GET_*" ( like webpy autodelegate + inputAsArgs able)
  - manage 404 thru "Web.page_notfound()"
  - Web.run()/Web.app() method to run the route's app (middlewar'able)
  - cookies simplified ( +set_cookie(), + del_cookie() )
  - easy get input parameters (instance act as a dict) (one/multiple/file values)
  - unicode/utf8 aware
  - unittest'able easily
  - Web.is_dev
  - can be run outside of GAE (as a httpserver (need webapp/webob))
  - lite

v0.2 ... IN PROGESS ...
=======================================================================================

EXEMPLE :
    from aeweb import Web

    class MyApp(Web):
      def GET(self,n="world"):
        yield "Hello "
        yield n
        yield self["jo"]    # ?jo=hello

    def main():
      Web.run([('/', MyApp),('/(.*)', MyApp),], debug=True)

    if __name__ == '__main__':
      main()
"""
try:
  from google.appengine.ext import webapp
except: # to be able to be run from the main file in local
  import sys
  sys.path = [".."]+sys.path
  from google.appengine.ext import webapp

import wsgiref.handlers
import logging                  # pour log()
import Cookie,urllib            # pour Web.setCookie(), Web.delCookie()
from datetime import datetime   # pour Web.setCookie()
import new                      # for Web.app() (to add new instance methods)
import re                       # for Web.hostroute()
import os                       # for Web.is_dev

def log(m):
  logging.info(m)

def u8(v): # ensure valid string 'v' is UTF_8
  if v is not None:
    if type(v)==unicode:
      return v.encode("utf_8")
    else:
      return v
  else:
    return None

class Web(webapp.RequestHandler):

  is_dev=os.environ.get('SERVER_SOFTWARE','').startswith('Dev')

  @staticmethod
  def _trans( wsgiapp ):
    """ add 2 instance method to a wsgiapp """
    def _test(this,path,method="GET"): #,http_host=None):
      assert method in ["GET","POST","HEAD","PUT","DELETE"]
      r=webapp.Request.blank(path)
      #~ if http_host is not None:
        #~ r.environ["HTTP_HOST"] = http_host
      r.method=method
      return r.get_response(this)

    def _run( this,exceptHosts=None,serverPort=None):
      """ run the actual wsgi'app, except for htt_host defined in exceptHost
        exceptHost can be a dict : { wsgiapp1:"localhost:8000", wsgiapp2:["localhost:8001","localhost"] }
        define serverPort to use aeweb as standalone (need webapp+webob)
      """
      if exceptHosts is not None:
        # there is "app hosting", we will insert a app
        # which will redirect according http_host for these excepthosts
        def application(environ, start_response):
            hostEnv=environ.get("HTTP_HOST","")
            for wsgiapp,strOrList in exceptHosts.items():
              hosts=[strOrList,] if type(strOrList)!=list else strOrList
              for hostRegex in hosts:
                if re.match(hostRegex,hostEnv):
                  # found an exception, run it
                  return wsgiapp(environ,start_response)

            # don't find an exception, run the default wsgiapp
            return this(environ,start_response)

        appToRun = application
      else:
        appToRun = this

      if serverPort:    # use as standalone
        from wsgiref.simple_server import make_server
        print "HttpServer on localhost:%d" % serverPort
        make_server('',serverPort,appToRun).serve_forever()
      else:             # use in a "gae env" (as cgi) (default)
        #~ wsgiref.handlers.CGIHandler().run(appToRun)
        return appToRun

    wsgiapp.run  = new.instancemethod(_run ,wsgiapp,wsgiapp.__class__)
    wsgiapp.test = new.instancemethod(_test,wsgiapp,wsgiapp.__class__)

    return wsgiapp

  #~ @staticmethod
  #~ def apphost( routes ):
    #~ """ return a 'wsgi app' according host-'routes' (list of tuple (regex, wsgi app)) """
    #~ def application(environ, start_response):
        #~ host=environ["HTTP_HOST"]
        #~ for regex,wsgiapp in routes:
          #~ if re.match(regex,host):
            #~ return wsgiapp(environ,start_response)

        #~ start_response('404 Not Found', [('Content-Type', 'text/html')])
        #~ return '404 not found'

    #~ return Web._trans(application)

  @staticmethod
  def app(routes,middlewares=None,debug=False):
    """ return a 'wsgi app', according 'routes' (list of tuple (regex, aeweb's app))
        middleware can be a list of middlewares ;-)
        if debug, exceptions are returned as html traceback

        the "wsgi app" contains new 2 methods :
          - run() : which run the app in the gae way (cgihandler)
          - test(path,method='GET') : which return a response object for 'path'/'method'
    """
    application = webapp.WSGIApplication(routes,debug=debug)
    if middlewares:
      for m in middlewares:
        application=m(application)

    return Web._trans(application)


  @staticmethod
  def run(routes,middlewares=None,debug=False): #FOR OLD COMPATIBILITY
    """ run Web's app, according 'routes' (list of tuple (regex, aeweb's app))
        middleware can be a list of middlewares ;-)
        if debug, exceptions are returned as html traceback
    """
    Web.app(routes,middlewares,debug).run()


  def delegate(self,prefix,page,inputAsArgs=False):
    """ re-route to self method 'prefix+page' (to prefix+page(inputArgs) if inputAsArgs)
    if method doesn't exist, call self.notfound() !"""
    method=prefix+(page and page.replace("/","_") or "index")
    if hasattr(self,method):
      if inputAsArgs:
        args = dict([(u8(k),self[k]) for k in self.request.arguments()])
        try:
          return getattr(self,method)(**args)
        except TypeError:
          return self.notfound()
      else:
        return getattr(self,method)()
    else:
      return self.notfound()

  def notfound(self):               #TO BE CALLED !
      """ send a '404 Not Found' (use page_notfound() to generate content)"""
      self.error(404)
      return Web.page_notfound( self )

  @staticmethod
  def page_notfound( instance=None ):    # TO BE OVERRIDEN !
      """return the html content for a '404', override this to define your own content"""
      if instance:
        assert type(instance)==Web
        return "Not Found on %s " % instance.url
      else:
        return "Not found"

  @property
  def environ(self): return self.request.environ

  @property
  def url(self): return self.request.url

  @property
  def path(self): return self.request.path

  @property
  def cookies(self):
    """ get the dict of cookies (unicoded values) """
    d={}
    d.update(self.request.cookies)
    for k,v in d.iteritems():
      d[k] = urllib.unquote(u8(v)).decode("utf_8")
    return d

  def set_cookie(self,name, value, expire=None, domain=None, path="/",secure=False):
    """ set a cookie named 'name' with value 'value', with states domain/path/secure optionnals
        'expire' must be a datetime for expiration of the cookie, or None
    """
    kargs = {'path':u8(path)}
    if expire:
        assert type(expire) == datetime
        kargs["expires"]=expire.strftime("%d %b %Y %H:%M:%S GMT")
    if domain:
        kargs['domain'] = u8(domain)
    if secure:
        kargs['secure'] = secure
    cookie = Cookie.SimpleCookie()
    cookie[u8(name)] = urllib.quote(u8(value))
    for key, val in kargs.iteritems():
        cookie[name][key] = val
    self.response.headers.add_header( 'Set-Cookie', cookie.items()[0][1].OutputString() )

  def del_cookie(self,name):
    """ delete the cookie named 'name' """
    cookie = Cookie.SimpleCookie()
    cookie[u8(name)] = ""
    cookie[u8(name)]["expires"] = -1
    self.response.headers.add_header( 'Set-Cookie', cookie.items()[0][1].OutputString() )


  def has_key(self,p):
    """ test if the input parameter 'p' is present in the request """
    return u8(p) in map(u8,self.request.arguments())

  def __getitem__(self,p):
    """ get the content of the input parameter 'p' (magic method ;-( )
    - if p has multiple value  -> return list of unicode string
    - if p has one value       -> return unicode string
    - if p was a file uploaded -> return tuple (name,mime,content) (all str of encoded page)
    """
    p=u8(p)
    l=self.request.get_all(p)
    if len(l)==1:
      # one unique element
      val = l[0]
      if hasattr(self.request.body_file,"vars") and \
         (p in self.request.body_file.vars) and \
         hasattr(self.request.body_file.vars[p],"filename"):
        # a "file" was uploaded -> tuple (name, mime, content)
        return (self.request.body_file.vars[p].filename, self.request.body_file.vars[p].headers['content-type'], val )
      else:
        # return the value -> unicode
        return val
    else:
      # multiple elements -> list of unicode
      return l

  def header(self,n,v):
    """ add an header in the response """
    self.response.headers[u8(n)] = u8(v)


  # manage restful apis (get/post/put ..) pointing to GET/POST/PUT ...
  def _do(self,method,*a,**k):
    if hasattr(self,method):
      caller = getattr(self,method)
      ret = caller(*a,**k)
      if hasattr(ret,"next"):
        map(self.response.out.write, ret)
      else:
        self.response.out.write(ret)
    else:
      self.error(405)
      self.response.out.write("method not allowed")

  def get(self,*a,**k): self._do("GET",*a,**k)
  def post(self,*a,**k): self._do("POST",*a,**k)
  def put(self,*a,**k): self._do("PUT",*a,**k)
  def head(self,*a,**k): self._do("HEAD",*a,**k)
  def delete(self,*a,**k): self._do("DELETE",*a,**k)


